# Rebellion Mod
# 
# Originally by Jon 'Trip' Shafer - 05 Nov 2005
# Civilization 4 (c) 2005 Firaxis Games
#
# Adapted by jdog5000
# Version 0.6
#
# Have fun!

from CvPythonExtensions import *
import sys
import Popup as PyPopup
from PyHelpers import PyPlayer
import pickle
import CvEventManager
from PyHelpers import *
import CvUtil
from math import sqrt
from CvCameraControls import CvCameraControls
# For offering player control of rebels
import ChangePlayer

# globals
gc = CyGlobalContext()
localText = CyTranslator()
game = CyGame()

DefaultUnitAI = UnitAITypes.NO_UNITAI

class Rebellion :

	def __init__(self, customEM, config):

                CvUtil.pyPrint("Initializing Rebellion Mod")
		
                self.LOG_DEBUG = config.getboolean("Rebellion", "DebugMode", False)
                self.bPythonOnly = config.getboolean("Rebellion", "PythonOnlyMode", False)
        		
		####### System Variables #######
		
		# The City Script array IDs when loading/saving data
		self.iCitySpawnScriptID = 0
		self.iCityRebelPlayerID = 1
		
		####### Game Variables #######
		
		self.iBarbariansID = gc.getBARBARIAN_PLAYER()
		self.rebelOfferPopup = 7005
		# Have to store data for popup handling for human player ... better way?
		self.iBuyoffCost = 0
		self.pCity = None
		self.iTotalRebelValue = 0
		self.bDidOfferControl = False
		self.pRebel = None
		self.isJoinWar = False
		self.isJoinRebellion = False
		
		# Chance of Rebel unit spawning near a city while the civ is in a Revolution
		self.iRebelSpawnRoll = config.getint("Rebellion", "Roll", 10000) #10000
		# Range from cities rebel units can spawn
		self.iUnitSpawnRange = 2#2
		# Number of turns between when rebel units can spawn
		self.iCitySpawnUnitCounterMax = config.getint("Rebellion", "Turns", 10)#6
		# Threshold below which there is no chance of rebellion
		self.iThreshold = config.getint("Rebellion", "Threshold", 1200)
		self.iUnitThreshold = config.getint("Rebellion", "UnitThreshold", 400)
		# If true, cities with no unhappy citizens can also rebel
		self.allowHappyRebellions = config.getboolean("Rebellion", "AllowHappyRebellions", False)
		# If True, spawn rebels as Barbarians, otherwise try to spawn a new rebel player
		self.rebelsAsBarbs = config.getboolean("Rebellion", "RebelsAsBarbs", False)
		# Offer the human player control of a rebellion that occurs in their territory
		self.OfferControl = config.getboolean("Rebellion", "OfferControlOfRebels", True )
		
		############## CITY VS STATE RELIGION ##############
		
		# Modifies how much a city having a different religion from the state affects chance of rebels
		self.iDifferentReligionFactor = 150 # Default is 150
		# Increase different religion factor when running a theocracy
		self.iTheocracyMultiplier = 4
		# Increase different religion factor when running Organized religion
		self.iOrganizedReligionMultiplier = 2
		
		############## CIVICS ##############
				
		# Modifies how much a player's government civic option affects chance of rebels
		self.iDespotismFactor = 150 # Default is 50
		self.iHereditaryRuleFactor = 100 # Default is 50
		self.iPoliceStateFactor = 200 # Default is 100
		self.iUniversalSuffrageFactor = -50
		
		# Modifies how much a player's legal civic option affects chance of rebels
		self.iBarbarismFactor = 100 # Default is 50
		self.iVassalageFactor = 75 # Default is 75
		
		# Modifies how much a player's labor civic option affects chance of rebels
		self.iTribalismFactor = 100 # Default is 50
		self.iSlaveryFactor = 150 # Default is 100
		self.iSerfdomFactor = 100 # Default is 100
		self.iCasteSystemFactor = 85 # Default is 85
		
		# Modifies how much a player's religion civic option affects chance of rebels
		self.iPaganismFactor = 50 # Default is 50
		# Most of effect for these civics is factored into religion factor
		self.iOrganizedReligionFactor = 50#100 # Default is 100
		self.iTheocracyFactor = 100#100 # Default is 100
		
		############## LAND AREA ##############
		
		# Modifies how much a player's total amount of land affects chance of rebels
		self.fLandModifier = config.getfloat("Rebellion", "LandFactor", 1.00)
		
		############## HEALTH AND HAPPINESS ##############
		
		# Modifies how much a player's city unhealthiness affects chance of rebels
		self.iHealthFactor = 200 # Default is 200
		self.iHealthyDivisor = 8
		self.iHappinessFactor = 400 # Default is 300
		self.iHappyDivisor = 6
		# Multiplies total rebel value when city is starving
		self.starvingModifier = 1.5
		
		############## CITY NATIONALITY ##############
		
		self.iNationalityFactor = 1000 # Default is 1000
		
		############## CIV SIZE ##############
		
		# Modifies how much a player's total land area affects chance of rebels
		self.iCivSizeFactor = 500 # Default is 1000
		# When a tiny civ has a distant colony, this caps the size of the modifier
		self.maxCityDistModifier = 2.0

		############## CITY AREA ###############

		# Modifies how much city being in a different area than capital affects chance of rebels
		self.iCityAreaFactor = 300
		
		############## ? ##############
		
		# Modifies how much a  affects chance of rebels
		self.fModifier = config.getfloat("Rebellion", "Modifier", 1.00)
		self.fSizeModifier = config.getfloat("Rebellion", "RebellionSizeModifier", 1.00)
		
		self.szEventText = ""
		self.szResultText = ""

		self.customEM = customEM

		# Register Rebellion events:
		self.customEM.addEventHandler( "EndGameTurn", self.onEndGameTurn )
		self.customEM.addEventHandler( "cityBuilt", self.onCityBuilt )
		self.customEM.addEventHandler( "cityAcquired", self.onCityAcquired )

		self.customEM.setPopupHandler( self.rebelOfferPopup, ["rebelOfferPopup",self.rebelOfferHandler,self.blankHandler] )

	def blankHandler( self, playerID, netUserData, popupReturn ) :
            # Dummy handler to take the second event for popup
            return
			
	def turnChecker(self, iTurn):
		
		self.szGameDate = CyGameTextMgr().getTimeStr(iTurn, false)
		
		# Alter City Unit Spawning Counters
		self.alterCitySpawnUnitCounters()
		
		# Determine where rebels show up and whatnot
		self.checkForRebellion(iTurn)

		# Remove floating rebels (ie rebels at peace, but have no cities)
		self.removeFloatingRebellions()
		
###################### Events ####################################

	def onEndGameTurn(self, argsList):
		'Called at the end of the end of each turn'
		iGameTurn = argsList[0]
		
		# Everything that happens at the end of a turn pertaining to the mod comes through here
		self.turnChecker(iGameTurn)
		
	def onCityBuilt(self, argsList):
		'City Built'
		city = argsList[0]
		
		# Init city script data (unit spawn counter, rebel player)
		self.initCity(city)
		
	def onCityAcquired(self, argsList):
		'City Acquired'
		
		owner,playerType,city,bConquest,bTrade = argsList

		if( not bConquest ) :
                    # Init city script data (unit spawn counter, rebel player)
                    self.initCity(city)
                    return

		newOwnerID = city.getOwner()
		oldOwnerID = city.getPreviousOwner()
		orgOwnerID = city.getOriginalOwner()		

		if( newOwnerID == gc.getBARBARIAN_PLAYER() and self.getCitySpawnUnitCounter( oldOwnerID, city ) > 0 ) :
                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - City %s captured by barb rebels!"%(city.getName()))
                    oldOwner = gc.getPlayer(oldOwnerID)
                    
                    if( not oldOwnerID == orgOwnerID ) :
                        orgOwner = gc.getPlayer(orgOwnerID)
                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - City %s originally owned by %s!"%(city.getName(),orgOwner.getCivilizationDescription(0)))

                        #if( something ) :
                            # Flip back to original owner

                            # Kill nearby barbs                    

                    if( city.countTotalCulture() > 100 ) :
                        if( not (oldOwnerID == city.findHighestCulture()) ) :
                            cultOwner = gc.getPlayer( city.findHighestCulture() )
                            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - City %s culturally dominated by %s!"%(city.getName(),cultOwner.getCivilizationDescription(0)))

                            #if( something ) :
                                # Flip to cultural owner

                                # Gift nearby barbs

                elif( newOwnerID == self.getCityRebelPlayer(city) and self.getCitySpawnUnitCounter( oldOwnerID, city ) > 0 ) :
                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Rebellious city %s is captured by rebel player %d!!!"%(city.getName(),newOwnerID))

                    cultPlayer = None
                    if( city.countTotalCulture() > 100 ) :
                        if( not city.getPreviousOwner() == city.findHighestCulture() ) :
                            cultPlayer = gc.getPlayer( city.findHighestCulture() )
                            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - New rebel city %s culturally dominated by %"%(city.getName(),cultPlayer.getCivilizationDescription(0)))

                    if( city.plot().countTotalCulture() > 100 ) :
                        if( (not city.plot().calculateCulturalOwner() == city.getPreviousOwner()) and cultPlayer == None ) :
                            cultPlayer = gc.getPlayer( city.plot().calculateCulturalOwner() )
                            if( not cultPlayer == None ) :
                                if( cultPlayer.isBarbarian() ) :
                                    cultPlayer = None
                                else :
                                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - New rebel city %s is culturally influenced by %"%(city.getName(),cultPlayer.getCivilzationDescription(0)))

                    city.setCulture( newOwnerID, city.getCulture( city.getPreviousOwner() ), True )
                    city.plot().setCulture( newOwnerID, city.plot().getCulture(city.getPreviousOwner()), True )
                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Culture set to %s"%(city.getCulture(newOwnerID)))

                    ix = city.getX()
                    iy = city.getY()

                    newOwner = gc.getPlayer( newOwnerID )

                    if( newOwner.isHuman() ) :
                        CyInterface().addImmediateMessage("With your capture of %s, support for the rebellion in the city swells ..."%(city.getName()),"")
                    elif( city.isVisible( game.getActiveTeam(), False ) ) :
                        CyInterface().addImmediateMessage("The rebel %s have taken control of %s!  Support for the rebellion swells ..."%(newOwner.getCivilizationDescription(0),city.getName()),"")

                    workerClass = CvUtil.findInfoTypeNum(gc.getUnitClassInfo,gc.getNumUnitClassInfos(),'UNITCLASS_WORKER')
                    iWorker = gc.getCivilizationInfo( newOwner.getCivilizationType() ).getCivilizationUnits(workerClass)
                    iBestDefender = UnitTypes.NO_UNIT
                    iCounter = UnitTypes.NO_UNIT
                    iAttack = UnitTypes.NO_UNIT
                    for unitID in range(0,gc.getNumUnitInfos()) :
                        if( city.canTrain(unitID,False,False) and gc.getUnitInfo(unitID).getDomainType() == DomainTypes.DOMAIN_LAND ): # canTrain only does technology check
                            unitClass = gc.getUnitInfo(unitID).getUnitClassType()
                            # Defender (Archer,Longbow)
                            if( gc.getUnitInfo(unitID).getDefaultUnitAIType() == UnitAITypes.UNITAI_CITY_DEFENSE ):
                                if( (iBestDefender == UnitTypes.NO_UNIT) or gc.getUnitInfo(unitID).getCombat() >= gc.getUnitInfo(iBestDefender).getCombat() ) :
                                    #if( self.LOG_DEBUG ) : CvUtil.pyPrint("  BC - Civ def units return value: %d from %d, comp to %d"%(gc.getCivilizationInfo( newOwner.getCivilizationType() ).getCivilizationUnits(unitClass),unitClass,unitID))
                                    if( unitID == gc.getCivilizationInfo( newOwner.getCivilizationType() ).getCivilizationUnits(unitClass) ) :
                                        iBestDefender = unitID
                                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Best defender set to %s"%(PyInfo.UnitInfo(iBestDefender).getDescription()))
                            # Counter (Axemen,Phalanx)
                            if( gc.getUnitInfo(unitID).getDefaultUnitAIType() == UnitAITypes.UNITAI_COUNTER ):
                                if( (iCounter == UnitTypes.NO_UNIT) or gc.getUnitInfo(unitID).getCombat() >= gc.getUnitInfo(iCounter).getCombat() ) :
                                    #if( self.LOG_DEBUG ) : CvUtil.pyPrint("  BC - Civ counter units return value: %d from %d, comp to %d"%(gc.getCivilizationInfo( newOwner.getCivilizationType() ).getCivilizationUnits(unitClass),unitClass,unitID))
                                    if( unitID == gc.getCivilizationInfo( newOwner.getCivilizationType() ).getCivilizationUnits(unitClass) ) :
                                        iCounter = unitID
                                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Best counter unit set to %s"%(PyInfo.UnitInfo(iCounter).getDescription()))
                            # Assault units
                            if( gc.getUnitInfo(unitID).getUnitAIType( UnitAITypes.UNITAI_ATTACK ) ):
                                if( (iAttack == UnitTypes.NO_UNIT) or gc.getUnitInfo(unitID).getCombat() > gc.getUnitInfo(iAttack).getCombat() ) :
                                    #if( self.LOG_DEBUG ) : CvUtil.pyPrint("  BC - Civ attack units return value: %d from %d, comp to %d"%(gc.getCivilizationInfo( newOwner.getCivilizationType() ).getCivilizationUnits(unitClass),unitClass,unitID))
                                    if( unitID == gc.getCivilizationInfo( newOwner.getCivilizationType() ).getCivilizationUnits(unitClass) ) :
                                        iAttack = unitID
                                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Best attack set to %s"%(PyInfo.UnitInfo(iAttack).getDescription()))

                    if( not iBestDefender == UnitTypes.NO_UNIT ) : newOwner.initUnit( iBestDefender, ix, iy, UnitAITypes.NO_UNITAI )
                    if( not iBestDefender == UnitTypes.NO_UNIT ) : newOwner.initUnit( iBestDefender, ix, iy, UnitAITypes.NO_UNITAI )
                    if( not iCounter == UnitTypes.NO_UNIT ) : newOwner.initUnit( iCounter, ix, iy, UnitAITypes.NO_UNITAI )
                    newOwner.initUnit( iWorker, ix, iy, UnitAITypes.NO_UNITAI )
                    if( gc.getPlayer( newOwnerID ).getNumCities() <= 1 ) :
                        # Extra units for first city captured
                        if( not iAttack == UnitTypes.NO_UNIT ) : newOwner.initUnit( iAttack, ix, iy, UnitAITypes.NO_UNITAI )
                        if( not iAttack == UnitTypes.NO_UNIT ) : newOwner.initUnit( iAttack, ix, iy, UnitAITypes.NO_UNITAI )
		
		# Init city script data (unit spawn counter, rebel player)
		self.initCity(city)
		iTurns = city.getOccupationTimer()
		#if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Occupations timer is currently %d"%(iTurns))
		self.setCitySpawnUnitCounter(newOwnerID, city, int(1.5*iTurns))

###########################################################################################
####################################### REBEL EVENTS ######################################
###########################################################################################
	
	def checkForRebellion(self, iTurn):
		
		# Loop through all players...
		for iPlayerID in range(gc.getMAX_CIV_PLAYERS()):
			
			pPlayer = gc.getPlayer(iPlayerID)
#			pyPlayer = PyPlayer(iPlayerID)
			
			# If player is alive...
			if (pPlayer.isAlive()):
				
########################## This is where the fun begins ##########################
				
				# This player's state religion
				iStateReligionID = pPlayer.getStateReligion()
				
				# Civics
				iCivicsValue = self.getCivicsFactor(pPlayer)
				
				# Civ size in land plots
				iCivNumLandPlots = pPlayer.getTotalLand()
				
########################## Loop through player's cities... ##########################
				
				apCityList = self.getCityList(pPlayer)
				for pCity in apCityList:

                                    unhappyPop = (pCity.unhappyLevel(0) - pCity.happyLevel())

                                    if( pCity.isOccupation() ) :
                                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - City %s is under occupation by %s" %(pCity.getName(),pPlayer.getCivilizationDescription(0)))
                                    elif( unhappyPop > 0 or self.allowHappyRebellions ) :
					
					# City factors
					
					# City religions vs. state religion
					iReligionValue = self.getReligionFactor(iStateReligionID, pCity)
					# Modify religion value based on civics
					iCivicOptionReligion = CvUtil.findInfoTypeNum(gc.getCivicOptionInfo, gc.getNumCivicOptionInfos(), 'CIVICOPTION_RELIGION')
                                        iReligionCivic = pPlayer.getCivics(iCivicOptionReligion)
                                        iOrganizedReligionID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_ORGANIZED_RELIGION')
                                        iTheocracyID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_THEOCRACY')
                                        if( iReligionCivic == iOrganizedReligionID ) :
                                            iReligionValue = iReligionValue*self.iOrganizedReligionMultiplier
                                        elif( iReligionCivic == iTheocracyID ) :
                                            iReligionValue = iReligionValue*self.iTheocracyMultiplier
					# Make religion less important after the game is nearing completion
					iHalfTurnNum = (CyGame().getMaxTurns() - CyGame().getStartTurn()) / 2
					if (CyGame().getGameTurn() > iHalfTurnNum):
						iTurnsAfterHalf = CyGame().getGameTurn() - iHalfTurnNum
						fReligionMod = iTurnsAfterHalf / (iHalfTurnNum * 1.0)
						iReligionValue -= iReligionValue * fReligionMod
					
					# Health
					numUnhealthy = (pCity.badHealth(false) - pCity.goodHealth())
					if( numUnhealthy <= 0 ) :
                                            iHealthValue = 0
                                        else :
                                            iHealthValue = int( pow(numUnhealthy, .7) * self.iHealthFactor )
										
					# Happiness
					iHappinessValue = int( pow(pCity.unhappyLevel(0) - pCity.happyLevel(),0.7) * self.iHappinessFactor )
					if (iHappinessValue < 0):
                                                # Happy people will over look SOME of your other problems
						iHappinessValue = iHappinessValue/self.iHappyDivisor
					else :
                                            # Reduce factor for capital
                                            if( pCity.isCapital() ) :
                                                iHappinessValue -= self.iHappinessFactor
                                                
                                            # Reduce happiness factor when civ is at war, people unhappy cause of war won't take up arms
                                            warCount = gc.getTeam(pPlayer.getTeam()).getAtWarCount(True)
                                            if( warCount > 0 ) :
                                                iHappinessValue = iHappinessValue/(warCount + 0.2)
					
					# City nationality
					iNationalityValue = (100 - pCity.plot().calculateCulturePercent(iPlayerID)) * self.iNationalityFactor / 100.0
					if( pCity.plot().countTotalCulture() < 100 ) :
                                                # If no civ has much culture here, reduce factor
                                                iNationalityValue = int(iNationalityValue*(pCity.countTotalCulture()/100.0))
                                        
##                                        if( iNationalityValue > 0 ) :
##                                                print "Rebel - Culture percent in %s: %d, plot: %d"%(pCity.getName(),pCity.calculateCulturePercent(iPlayerID),pCity.plot().calculateCulturePercent(iPlayerID))
##					print("XXX - Player %d Culture Percent: %d; (100 - %d) * %d" %(iPlayerID, pCity.calculateCulturePercent(iPlayerID), pCity.calculateCulturePercent(iPlayerID), self.iNationalityFactor))
					
					# Civ size
					defNumPlayers = gc.getWorldInfo(gc.getMap().getWorldSize()).getDefaultPlayers()
					expCivPlots = CyMap().getLandPlots() / ( defNumPlayers * 1.0)
					iCivSizeValue = iCivNumLandPlots / expCivPlots * self.iCivSizeFactor
                                        pCapital = pPlayer.getCapitalCity()
                                        # Distance to capital divided by something a little bigger than civ's "radius"
                                        cityDistModifier = plotDistance( pCity.getX(), pCity.getY(), pCapital.getX(), pCapital.getY() )/(sqrt(iCivNumLandPlots/3.0))
                                        oCityDistMod = cityDistModifier
                                        cityDistModifier = min([cityDistModifier, self.maxCityDistModifier])
                                        #if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - City %s has dist mod %f from %f"%(pCity.getName(),cityDistModifier,oCityDistMod))
                                        # Total civ size effect: some effect for all cities, more for distant cities
                                        iCivSizeValue = int( self.fLandModifier*(iCivSizeValue/2.0 + cityDistModifier*iCivSizeValue/2.0) )

					# City area
					iCityAreaValue = 0
					if( not (pCity.area().getID() == pPlayer.getCapitalCity().area().getID()) ) :
                                                iCityAreaValue += self.iCityAreaFactor

                                        # Some factors not as large pre-Nationalism
                                        iNationalismTech = CvUtil.findInfoTypeNum(gc.getTechInfo,gc.getNumTechInfos(),'TECH_NATIONALISM')
                                        if( not gc.getTeam(pPlayer.getTeam()).isHasTech(iNationalismTech) ) :
                                                #if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Not yet to nationalism")
                                                iNationalityValue = iNationalityValue/2
                                                iCityAreaValue = iCityAreaValue/2
					
					# Sum all factors
					iTotalRebelValue = iReligionValue + iCivicsValue + iHealthValue + iHappinessValue + iNationalityValue + iCivSizeValue
                                        #if( self.LOG_DEBUG ) : CvUtil.pyPrint("XXX - City %s has iTotalRebelValue of %d - Civics %d, Religion %d, Health %d, Happiness %d, Nationality %d, CivSize %d, CityArea %d" %(pCity.getName(), iTotalRebelValue, iCivicsValue, iReligionValue, iHealthValue, iHappinessValue, iNationalityValue, iCivSizeValue, iCityAreaValue))
                                        # All problems more severe when starving!
                                        if( pCity.foodDifference(True) < 0 and abs(pCity.foodDifference(True)) > pCity.getFood() ) :
                                            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - City %s is starving!"%(pCity.getName()))
                                            iTotalRebelValue = iTotalRebelValue*(self.starvingModifier)

                                        iTotalRebelValue = int( iTotalRebelValue )
				
                                        if( iTotalRebelValue > self.iThreshold ) :
                                            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - City %s has iTotalRebelValue of %d - Civics %d, Religion %d, Health %d, Happiness %d, Nationality %d, CivSize %d, CityArea %d" %(pCity.getName(), iTotalRebelValue, iCivicsValue, iReligionValue, iHealthValue, iHappinessValue, iNationalityValue, iCivSizeValue, iCityAreaValue))
					
                                            # Now spawn units
                                            if( self.doRebellion(iPlayerID, pCity, iTotalRebelValue) ) :
                                                # Only one rebellion per turn per player?
                                                break

	
	def getReligionFactor(self, iStateReligionID, pCity):
		
		iReligionFactor = 0
		
		# Loop through all religions
		for iCityReligionID in range(gc.getNumReligionInfos()):
			
			# If city has this religion...
			if (pCity.isHasReligion(iCityReligionID)):
				
				# If this city religion doesn't match the state religion...
				if (iStateReligionID != iCityReligionID):
					
					iReligionFactor += self.iDifferentReligionFactor
				else :
                                        iReligionFactor -= self.iDifferentReligionFactor/2
		
		return iReligionFactor
	
	def getCivicsFactor(self, pPlayer):
		
		# Government Civics
		iCivicOptionGovernment = CvUtil.findInfoTypeNum(gc.getCivicOptionInfo, gc.getNumCivicOptionInfos(), 'CIVICOPTION_GOVERNMENT')
		iGovernmentCivic = pPlayer.getCivics(iCivicOptionGovernment)
		
		iDespotismID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_DESPOTISM')
		iHereditaryRuleID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_HEREDITARY_RULE')
		iPoliceStateID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_POLICE_STATE')
		iUniversalSuffrageID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_UNIVERSAL_SUFFRAGE')
		
		# Legal Civics
		iCivicOptionLegal = CvUtil.findInfoTypeNum(gc.getCivicOptionInfo, gc.getNumCivicOptionInfos(), 'CIVICOPTION_LEGAL')
		iLegalCivic = pPlayer.getCivics(iCivicOptionLegal)
		
		iBarbarismID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_BARBARISM')
		iVassalageID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_VASSALAGE')
		
		# Labor Civics
		iCivicOptionLabor = CvUtil.findInfoTypeNum(gc.getCivicOptionInfo, gc.getNumCivicOptionInfos(), 'CIVICOPTION_LABOR')
		iLaborCivic = pPlayer.getCivics(iCivicOptionLabor)
		
		iTribalismID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_TRIBALISM')
		iSlaveryID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_SLAVERY')
		iSerfdomID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_SERFDOM')
		iCasteSystemID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_CASTE_SYSTEM')
		
		# Religion Civics
		iCivicOptionReligion = CvUtil.findInfoTypeNum(gc.getCivicOptionInfo, gc.getNumCivicOptionInfos(), 'CIVICOPTION_RELIGION')
		iReligionCivic = pPlayer.getCivics(iCivicOptionReligion)
		
		iPaganismID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_PAGANISM')
		iOrganizedReligionID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_ORGANIZED_RELIGION')
		iTheocracyID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_THEOCRACY')
		
		########## CALCULATE VALUES ##########
		
		iCivicValue = 0
		
		if (iGovernmentCivic == iDespotismID):
			iCivicValue += self.iDespotismFactor
		elif (iGovernmentCivic == iHereditaryRuleID):
			iCivicValue += self.iHereditaryRuleFactor
		elif (iGovernmentCivic == iPoliceStateID):
			iCivicValue += self.iPoliceStateFactor
		elif ( iGovernmentCivic ==  iUniversalSuffrageID ) :
                        iCivicValue += self.iUniversalSuffrageFactor
			
		if (iLegalCivic == iBarbarismID):
			iCivicValue += self.iBarbarismFactor
		elif (iLegalCivic == iVassalageID):
			iCivicValue += self.iVassalageFactor
			
		if (iLaborCivic == iTribalismID):
			iCivicValue += self.iTribalismFactor
		elif (iLaborCivic == iSlaveryID):
			iCivicValue += self.iSlaveryFactor
		elif (iLaborCivic == iSerfdomID):
			iCivicValue += self.iSerfdomFactor
		elif (iLaborCivic == iCasteSystemID):
			iCivicValue += self.iCasteSystemFactor
			
		if (iReligionCivic == iPaganismID):
			iCivicValue += self.iPaganismFactor
		elif (iReligionCivic == iOrganizedReligionID):
			iCivicValue += self.iOrganizedReligionFactor
		elif (iReligionCivic == iTheocracyID):
			iCivicValue += self.iTheocracyFactor
		
		return iCivicValue

###########################################################################################
####################################### OTHER EVENTS ######################################
###########################################################################################
	
	def doRebellion(self, iPlayerID, pCity, iTotalRebelValue):

            iCityX = pCity.getX()
	    iCityY = pCity.getY()
		
            # If this city has spawned a unit recently then no more units
            if (self.getCitySpawnUnitCounter(iPlayerID, pCity) == 0):
                    
                iUnitSpawnRoll = self.getRand(self.iRebelSpawnRoll)

                if (iUnitSpawnRoll < int(self.fModifier*iTotalRebelValue)):
                    # City is rebelling!
                    self.rebelOffer( iPlayerID, pCity, iTotalRebelValue )
                    return True

            else :
                if( self.LOG_DEBUG ) : print ("Rebel - City %s just revolted"%(pCity.getName()))

            return False

        

        def rebelOffer( self, iPlayerID, pCity, iTotalRebelValue ) :

            pPlayer = gc.getPlayer( iPlayerID )
            iEra = pPlayer.getCurrentEra()
            iRandCost = 300 - self.getRand( 400 )
            iBuyoffCost = int( (iTotalRebelValue + iRandCost)/(4+gc.getNumEraInfos()-iEra) )

            [pRebel, isJoinWar, isJoinRebellion] = self.chooseRebelCiv( iPlayerID, pCity )
            bIsBarb = pRebel.isBarbarian()

            if( isJoinWar ) :
                # City won't have many soldiers as it's probably been fought over for a while
                iBuyoffCost = 2*iBuyoffCost/3

            if( pPlayer.isHuman() ) :
                # Store variables for handler function
                self.iBuyoffCost = iBuyoffCost
                self.pCity = pCity
                self.iTotalRebelValue = iTotalRebelValue
                self.pRebel = pRebel
                self.isJoinWar = isJoinWar
                self.isJoinRebellion = isJoinRebellion

                # Center camera on city
                CyCamera().JustLookAt( pCity.plot().getPoint() )
                
                if( self.LOG_DEBUG ) : print ("Rebel - Buy off cost for this rebellion is %d"%(self.iBuyoffCost))
                
                # Offer human player the chance to negotiate with Rebels
                popup = PyPopup.PyPopup(self.rebelOfferPopup,contextType = EventContextTypes.EVENTCONTEXT_ALL)
                popup.setHeaderString( 'A Rebellion is Brewing!' )
                bodStr = ''
                if( bIsBarb ) :
                    bodStr += 'A leader from the rebellious city of %s approaches you with an offer.\n\n'%(pCity.getName())
                elif( isJoinWar ) :
                    bodStr += 'A leader from the %s insurgency in the city of %s approaches you with an offer.\n\n'%(pRebel.getCivilizationAdjective(0),pCity.getName())
                else :
                    bodStr += 'The %s leader from the the rebellious city of %s approaches you with an offer.\n\n'%(pRebel.getCivilizationAdjective(0),pCity.getName())
                    
                bodStr += 'They would be willing to put down their arms and wait to see if conditions improve for the sum of %d.'%(self.iBuyoffCost)
                if( pPlayer.getGold() < self.iBuyoffCost ) :
                    bodStr += '  Unfortunately, you do not have enough in the treasury.'
                self.bDidOfferControl = False
                if( not bIsBarb and not isJoinWar and not self.bPythonOnly and self.OfferControl ) :
                    self.bDidOfferControl = True
                    bodStr += "\n\nWhile the rebel leader will only speak in innuendo in front of your advisors, "
                    bodStr += "you get the feeling the %s would welcome you as their new leader."%(pRebel.getCivilizationAdjective(0))
                    #bodStr += "  You would have to forsake all that you have built and face your own armies on the field."
                    bodStr += "  (Note: if you take this option, you will become %s of the %s)"%(pRebel.getName(),pRebel.getCivilizationDescription(0))
                popup.setBodyString( bodStr )

                popup.addButton('The %s does not negotiate with rebels.'%(pPlayer.getCivilizationDescription(0)))
                if( self.bDidOfferControl ) :
                    popup.addButton("Rebellion?  Count me in!")
                if( pPlayer.getGold() > self.iBuyoffCost ) :
                    popup.addButton('We can afford the %d gold.'%(self.iBuyoffCost))
                #popup.addPythonButton( 'Rebellion.doBuyoff', 'We can afford the %d gold.'%(self.iBuyoffCost), 'Will buyoff the rebels, putting off the rebellion for some time', iData1 = iPlayerID, iData2 = pCity.getID() )

                popup.launch(bCreateOkButton = False)
                
                return

            else :
                # AI player's city is rebelling, see if they can buy them off
                iBuyoffCost = 2*iBuyoffCost/3
                if( self.LOG_DEBUG ) : print ("Rebel - AI Buy off cost for this rebellion is %d"%(iBuyoffCost))
                
                if( pPlayer.getGold() >= iBuyoffCost ) :
                    iAgg = CvUtil.findInfoTypeNum(gc.getTraitInfo,gc.getNumTraitInfos(),'TRAIT_AGGRESSIVE')
                    if( pPlayer.hasTrait( iAgg ) ) :
                        base = .85
                    else :
                        base = .7

                    oddsForBuy = 1.0 - pow( base, pPlayer.getGold()/(1.0*iBuyoffCost) )
                    oddsForBuy = int(100*oddsForBuy)

                    if( self.LOG_DEBUG ) : print ("Rebel - Odds for buyoff are %d, with base %f"%(oddsForBuy,base))

                    if( oddsForBuy >= self.getRand( 100 ) ) :
                        if( self.LOG_DEBUG ) : print ("Rebel - AI is buying off rebels")
                        pPlayer.changeGold( -iBuyoffCost )
                        self.setCitySpawnUnitCounter(iPlayerID, pCity, self.iCitySpawnUnitCounterMax)
                    elif( oddsForBuy >= 40 and gc.getTeam(pPlayer.getTeam()).getAtWarCount(True) > 1 ) :
                        if( self.LOG_DEBUG ) : print ("Rebel - AI is in multiple wars, desperately buying off")
                        pPlayer.changeGold( -iBuyoffCost )
                        self.setCitySpawnUnitCounter(iPlayerID, pCity, self.iCitySpawnUnitCounterMax)
                    else :
                        if( self.LOG_DEBUG ) : print ("Rebel - AI opted not to buy off rebels")
                        self.spawnRebels( iPlayerID, pCity, iTotalRebelValue, pRebel, isJoinWar, isJoinRebellion )
                    
                else :
                    if( self.LOG_DEBUG ) : print ("Rebel - AI doesn't have the money")
                    self.spawnRebels( iPlayerID, pCity, iTotalRebelValue, pRebel, isJoinWar, isJoinRebellion )

        def rebelOfferHandler( self, iPlayerID, netUserData, popupReturn ) :
           
            if( popupReturn.getButtonClicked() == 0 ) :
                if( self.LOG_DEBUG ) : print ("Rebel - Clicked option 0, no deal")

                self.spawnRebels( iPlayerID, self.pCity, self.iTotalRebelValue, self.pRebel, self.isJoinWar, self.isJoinRebellion )

            elif( popupReturn.getButtonClicked() == 1 ) :
                if( self.bDidOfferControl ) :
                    if( self.LOG_DEBUG ) : print ("Rebel - Clicked option 1, taking control of rebellion!")

                    # Human gets a little extra boost
                    rebelID = self.spawnRebels( iPlayerID, self.pCity, self.iTotalRebelValue + 2*self.iUnitThreshold, self.pRebel, self.isJoinWar, self.isJoinRebellion )

                    try :
                        success = ChangePlayer.changeHuman( rebelID, iPlayerID )
                    except AttributeError:
                        success = False
                    #success = ChangePlayer.changeHuman( rebelID, iPlayerID )

                    if( success ) :
                        if( self.LOG_DEBUG ) : print ("Rebel - Human now controls rebellion ...")
                    else :
                        if( self.LOG_DEBUG ) : print ("Rebel - ERROR: attempted to switch human from %d to %d failed"%(iPlayerID,rebelID))
                        popup = PyPopup.PyPopup()
                        popup.setHeaderString( 'An error has occured ...' )
                        bodStr = 'My apologies, but your attempt to take control of the rebellion has failed.  '
                        bodStr += 'Fortunately your people are unaware of your attempts to betray them, '
                        bodStr += 'so you retain control of your empire.'
                        popup.setBodyString( bodStr )
                        popup.launch()                        

                else :
                    if( self.LOG_DEBUG ) : print ("Rebel - Clicked option 1, buyoff")

                    gc.getPlayer( iPlayerID ).changeGold( -self.iBuyoffCost )

                    self.setCitySpawnUnitCounter( iPlayerID, self.pCity, self.iCitySpawnUnitCounterMax)
            elif( popupReturn.getButtonClicked() == 2 ) :
                if( self.LOG_DEBUG ) : print ("Rebel - Clicked option 2, buyoff")

                gc.getPlayer( iPlayerID ).changeGold( -self.iBuyoffCost )

                self.setCitySpawnUnitCounter( iPlayerID, self.pCity, self.iCitySpawnUnitCounterMax)

            else :
                print ("Rebel - Unexpected return from rebel offer popup")
                self.spawnRebels( iPlayerID, self.pCity, self.iTotalRebelValue, self.pRebel, self.isJoinWar, self.isJoinRebellion )

        def chooseRebelCiv( self, iPlayerID, pCity ) :

                pRebel = None
                owner = gc.getPlayer( pCity.getOwner() )

                isJoinWar = False
                isJoinRebellion = False
                
                if( self.rebelsAsBarbs ) :
                    pRebel = gc.getPlayer(self.iBarbariansID)
                else :
                    # Is dominant culture at war with city owner?
                    if( pCity.countTotalCulture() > 50 ) :
                        cultPlayer = gc.getPlayer( pCity.findHighestCulture() )
                        cultTeam = gc.getTeam( cultPlayer.getTeam() )
                        if( cultTeam.isAtWar(owner.getTeam()) ) :
                            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel:  City %s is joining dom culture player %d (%s) in war against owner"%(pCity.getName(),cultPlayer.getID(),cultPlayer.getCivilizationDescription(0)))
                            if( game.getGameTurn() - pCity.getGameTurnAcquired() < 25 ) :
                                isJoinWar = True
                            else :
                                isJoinRebellion = True
                            pRebel = cultPlayer

                    if( pRebel == None ) :
                        # Is there a dead civ in this city that can rise from the ashes?
                        for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                            if( gc.getPlayer(i).isEverAlive() and (gc.getPlayer(i).getNumUnits() == 0) ) :
                                if( pCity.getCulture( i ) > 10 ) :
                                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel:  Dead player %d has culture in rebelling %s, respawning"%(i,pCity.getName()))
                                    pRebel = gc.getPlayer(i)
                                    self.giveRebelTechs( pRebel, gc.getPlayer(pCity.getOwner()) )
                                    break

                    if( pRebel == None ) :
                        # Is there another rebellion we can join?
                        for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                            # Must have units in this area (colonial rebels don't join homeland rebels)
                            if( pCity.area().getUnitsPerPlayer(i) > 0 ) :
                                teami = gc.getTeam( gc.getPlayer(i).getTeam() )
                                if( teami.isAtWar(owner.getTeam()) ) :
                                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel:  Player %d is at war with owner of %s"%(i,pCity.getName()))
                                    if( not gc.getPlayer(i).isFoundedFirstCity() ) :  # Is a homeless rebel
                                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel:  Player %d (homeless rebel) is rebelling against owner of %s!!!  Joining rebellion."%(i,pCity.getName()))
                                        pRebel = gc.getPlayer(i)
                                        isJoinRebellion = True
                                        break
                                    elif( gc.getPlayer(i).getCitiesLost() < 3 and gc.getPlayer(i).getNumCities() < 4 ) :
                                        if( game.getGameTurn() - gc.getPlayer(i).getCapitalCity().getGameTurnAcquired() < 30 ) :
                                            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel:  Player %d (young rebel) is rebelling against owner of %s!!!  Joining rebellion."%(i,pCity.getName()))
                                            pRebel = gc.getPlayer(i)
                                            isJoinRebellion = True
                                            break
                       

                    if( pRebel == None ) :
                        # Create a new rebellion (or return barb if not possible)
                        try :
                            pRebel = self.spawnRebelPlayer( pCity )
                        except :
                            if( self.LOG_DEBUG ) : print ("***Rebel - DLL not found, spawning as Barbarians***")
                            pRebel = gc.getPlayer(self.iBarbariansID)

                return [pRebel, isJoinWar, isJoinRebellion]
					
	def spawnRebels(self, iPlayerID, pCity, iTotalRebelValue, pRebel = None, isJoinWar = False, isJoinRebellion = False):

                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel:  A rebellion is occuring in %s in year %d"%(pCity.getName(),game.getGameTurnYear()))

		iCityX = pCity.getX()
		iCityY = pCity.getY()
					
                owner = gc.getPlayer( pCity.getOwner() )
##                isJoinWar = False
##                isJoinRebellion = False
##                
##                if( self.rebelsAsBarbs ) :
##                    pRebel = gc.getPlayer(self.iBarbariansID)
##                else :
##                    # Is dominant culture at war with city owner?
##                    if( pCity.countTotalCulture() > 50 ) :
##                        cultPlayer = gc.getPlayer( pCity.findHighestCulture() )
##                        cultTeam = gc.getTeam( cultPlayer.getTeam() )
##                        if( cultTeam.isAtWar(owner.getTeam()) ) :
##                            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel:  City %s is joining dom culture player %d in war against owner"%(pCity.getName(),cultPlayer.getID()))
##                            isJoinWar = True
##                            pRebel = cultPlayer
##                    
##
##                    if( pRebel == None ) :
##                        # Is there a dead civ in this city that can rise from the ashes?
##                        for i in range(0,gc.getMAX_CIV_PLAYERS()) :
##                            if( gc.getPlayer(i).isEverAlive() and (gc.getPlayer(i).getNumUnits() == 0) ) :
##                                if( pCity.getCulture( i ) > 10 ) :
##                                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel:  Dead player %d has culture in rebelling %s, respawning"%(i,pCity.getName()))
##                                    pRebel = gc.getPlayer(i)
##                                    self.giveRebelTechs( pRebel, gc.getPlayer(pCity.getOwner()) )
##                                    break
##
##                    if( pRebel == None ) :
##                        # Is there another rebellion we can join?
##                        for i in range(0,gc.getMAX_CIV_PLAYERS()) :
##                            # Must have units in this area (colonial rebels don't join homeland rebels)
##                            if( pCity.area().getUnitsPerPlayer(i) > 0 ) :
##                                teami = gc.getTeam( gc.getPlayer(i).getTeam() )
##                                if( teami.isAtWar(owner.getTeam()) ) :
##                                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel:  Player %d is at war with owner of %s"%(i,pCity.getName()))
##                                    if( not gc.getPlayer(i).isFoundedFirstCity() ) :  # Is a homeless rebel
##                                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel:  Player %d (homeless rebel) is rebelling against owner of %s!!!  Joining rebellion."%(i,pCity.getName()))
##                                        pRebel = gc.getPlayer(i)
##                                        isJoinRebellion = True
##                                        break
##                                    elif( gc.getPlayer(i).getCitiesLost() < 3 and gc.getPlayer(i).getNumCities() < 4 ) :
##                                        if( game.getGameTurn() - gc.getPlayer(i).getCapitalCity().getGameTurnAcquired() < 30 ) :
##                                            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel:  Player %d (young rebel) is rebelling against owner of %s!!!  Joining rebellion."%(i,pCity.getName()))
##                                            pRebel = gc.getPlayer(i)
##                                            isJoinRebellion = True
##                                            break
                       

                if( pRebel == None ) :
                    # Create a new rebellion (or return barb if not possible)
                    try :
                        pRebel = self.spawnRebelPlayer( pCity )
                    except :
                        if( self.LOG_DEBUG ) : print ("***Rebel - DLL not found, spawning as Barbarians***")
                        pRebel = gc.getPlayer(self.iBarbariansID)
                        
                iTurns = min(iTotalRebelValue/500,6)
                pCity.setOccupationTimer( iTurns )

                bIsBarb = pRebel.isBarbarian()

                if( not bIsBarb and not isJoinWar ) :
                    if( not isJoinRebellion ) :
                        pRebel.setAlive( True )
                        
                        # Declare war on city owner
                        gc.getTeam( pRebel.getTeam() ).declareWar( pCity.getOwner(), True )
                        if( self.LOG_DEBUG ) : print ("Rebel - Rebel civ %s has declared war on %s" %(pRebel.getCivilizationDescription(0),gc.getPlayer(pCity.getOwner()).getCivilizationDescription(0)))

                    # Set city's rebel player
                    self.setCityRebelPlayer( pRebel.getID(), pCity )


                aiSpawnPlot = self.findUnitPlacementPlot(self.iBarbariansID, iCityX, iCityY, True, False, True, iRange = self.iUnitSpawnRange, plotOwnerID = pCity.getOwner())
#				print("Spawn Plot: %d, %d" %(aiSpawnPlot[0], aiSpawnPlot[1]))
                if (aiSpawnPlot == "Oh Snap"):
                    self.setCitySpawnUnitCounter(iPlayerID, pCity, self.iCitySpawnUnitCounterMax/2)
                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel:  No spawnable plot!!!  Aborting Rebellion")
                    return

                spawnableUnits = list()
                trainableUnits = list()

                ownerPy = PyPlayer( owner.getID() )
                enemyPy = None
                if( isJoinWar or isJoinRebellion ) :
                    enemyPy = PyPlayer( pRebel.getID() )

                for unitID in range(0,gc.getNumUnitInfos()) :
                    ownerUnits = ownerPy.getUnitsOfType( unitID )
                    unitInfo = gc.getUnitInfo(unitID)
                    if( len( ownerUnits ) > 0 ) :
                        if( ownerUnits[0].canAttack() ) :
                            if( unitInfo.getDomainType() == DomainTypes.DOMAIN_LAND and ( unitInfo.getUnitAIType(UnitAITypes.UNITAI_ATTACK) or unitInfo.getUnitAIType(UnitAITypes.UNITAI_COUNTER) )):
                                unitClass = gc.getUnitInfo(unitID).getUnitClassType()
                                for unit in ownerUnits :
                                    if( plotDistance( unit.getX(), unit.getY(), pCity.getX(), pCity.getY() ) < 5 ) :
                                        spawnUnitID = gc.getCivilizationInfo( pRebel.getCivilizationType() ).getCivilizationUnits(unitClass)
                                        spawnableUnits.append( spawnUnitID )
                                        if( self.LOG_DEBUG ) : print "Rebel:  Can spawn from owner %s"%(PyInfo.UnitInfo(spawnUnitID).getDescription())
                                        if( spawnableUnits.count( spawnUnitID ) > 7 ) :
                                            break

                                for unit in ownerUnits :
                                    if( unit.area().getID() == pCity.area().getID() ):
                                        spawnUnitID = gc.getCivilizationInfo( pRebel.getCivilizationType() ).getCivilizationUnits(unitClass)
                                        spawnableUnits.append( spawnUnitID )
                                        if( self.LOG_DEBUG ) : print "Rebel:  Can spawn from Area %s"%(PyInfo.UnitInfo(spawnUnitID).getDescription())
                                        break

                    if( not enemyPy == None ) :
                        enemyUnits = enemyPy.getUnitsOfType( unitID )
                        if( len( enemyUnits ) > 0 ) :
                            if( enemyUnits[0].canAttack() ) :
                                if( unitInfo.getDomainType() == DomainTypes.DOMAIN_LAND and unitInfo.getUnitAIType( UnitAITypes.UNITAI_ATTACK )  ):
                                    unitClass = gc.getUnitInfo(unitID).getUnitClassType()
                                    for unit in enemyUnits :
                                        if( plotDistance( unit.getX(), unit.getY(), pCity.getX(), pCity.getY() ) < 7 ) :
                                            spawnUnitID = gc.getCivilizationInfo( pRebel.getCivilizationType() ).getCivilizationUnits(unitClass)
                                            spawnableUnits.append( spawnUnitID )
                                            if( self.LOG_DEBUG ) : print "Rebel:  Can spawn from enemy %s"%(PyInfo.UnitInfo(spawnUnitID).getDescription())

                    
                    if( pCity.canTrain(unitID,False,False) and gc.getUnitInfo(unitID).getDomainType() == DomainTypes.DOMAIN_LAND ):
                        if( gc.getUnitInfo(unitID).getUnitAIType( UnitAITypes.UNITAI_ATTACK ) ):
                            unitClass = gc.getUnitInfo(unitID).getUnitClassType()
                            if( bIsBarb ) :
                                domPlayerIdx = pCity.plot().calculateCulturalOwner()
                                domPlayer = gc.getPlayer(domPlayerIdx)
                                spawnUnitID = gc.getCivilizationInfo( domPlayer.getCivilizationType() ).getCivilizationUnits(unitClass)
                            else :
                                spawnUnitID = gc.getCivilizationInfo( pRebel.getCivilizationType() ).getCivilizationUnits(unitClass)
                                
                            if( not spawnUnitID in trainableUnits ) :
                                trainableUnits.append( spawnUnitID )
                                if( gc.getUnitInfo(unitID).getCombat() > 4 ) :
                                    trainableUnits.append( spawnUnitID )
                                    if( gc.getUnitInfo(unitID).getCombat() > 15 ) :
                                        trainableUnits.append( spawnUnitID )
                                if( self.LOG_DEBUG ) : print "Rebel:  Can build %s"%(PyInfo.UnitInfo(spawnUnitID).getDescription())

                if( len(spawnableUnits) < 1 ) :
                    spawnableUnits = trainableUnits

                if( len(spawnableUnits) < 1 ) :
                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel:  No spawnable units!!!  Aborting Rebellion")
                    return

                numUnits = int( self.fSizeModifier*(iTotalRebelValue + self.getRand(1000))/(1.0*self.iUnitThreshold) ) + 1
                if( isJoinWar ) :
                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Joining war, rebel enlistment limited") 
                    numUnits = min( [numUnits, pCity.getPopulation()/3] )
                else :
                    numUnits = min( [numUnits, pCity.getPopulation()] )

                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Spawning %d units for city of size %d"%(numUnits,pCity.getPopulation())) 

                for i in range(0,numUnits) :
                    newUnitID = spawnableUnits[self.getRand( len(spawnableUnits) )]
                                                        
                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Player %d in rebellion; Rebel %s spawning at [%d, %d]" %(iPlayerID,PyInfo.UnitInfo(newUnitID).getDescription(), aiSpawnPlot[0], aiSpawnPlot[1]))
                    pRebel.initUnit(newUnitID, aiSpawnPlot[0], aiSpawnPlot[1], UnitAITypes.UNITAI_ATTACK)

                if( gc.getPlayer(game.getActivePlayer()).canContact(iPlayerID) ) :
                    if( game.getActivePlayer() == iPlayerID ) :
                        mess = "A Rebellion has occured in %s!!!"%(pCity.getName())
                        if( bIsBarb ) :
                            mess += "  Disorganized rebel armies gather outside your gates!"
                        else :
                            mess += "  The %s have risen up against you!"%(pRebel.getCivilizationDescription(0)) 
                    else :
                        mess = "A Rebellion has occured in the %s!!!"%(PyPlayer(iPlayerID).getCivilizationName())
                        if( bIsBarb ) :
                            mess += "  Disorganized rebel armies gather in the countryside."
                        else :
                            mess += "  The %s have risen up to throw off their oppressors!"%(pRebel.getCivilizationDescription(0))
                    CyInterface().addImmediateMessage(mess,"")

                # Reveal surrounding territory
                iGoodyMap = CvUtil.findInfoTypeNum(gc.getGoodyInfo,gc.getNumGoodyInfos(),'GOODY_MAP')
                unitList = PyPlayer(pRebel.getID()).getUnitList()
                for unit in unitList :
                    if( unit.at(aiSpawnPlot[0],aiSpawnPlot[1]) ) :
                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Giving map")
                        pRebel.receiveGoody( gc.getMap().plot(aiSpawnPlot[0],aiSpawnPlot[1]), iGoodyMap, unit )
                        break

                # Set who can contact rebels
                for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                    if( pCity.plot().getCulture(i) > 0 ) :
                        if( not pRebel.canContact(i) ) :
                            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Rebel %d meeting %d"%(pRebel.getID(),i))
                            gc.getTeam( pRebel.getTeam() ).meet( gc.getPlayer(i).getTeam(), False )

                # Give a little gold
                pRebel.changeGold( numUnits*3 )
                                
                self.setCitySpawnUnitCounter(iPlayerID, pCity, self.iCitySpawnUnitCounterMax)

                return pRebel.getID()

        def spawnRebelPlayer( self, pCity, giveTechs = True ) :
            
            owner = gc.getPlayer( pCity.getOwner() )

            if( owner.getID() == self.iBarbariansID ) :
                return gc.getPlayer(self.iBarbariansID)
            
            cultOwner = None
            if( pCity.countTotalCulture() > 0 ) :
                cultOwner = gc.getPlayer( pCity.findHighestCulture() )
                if( cultOwner.getID() == self.iBarbariansID ) :
                    cultOwner = None

            if( cultOwner == None ) :
                cultOwner = owner

            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Spawning a rebel civ for the %s"%(cultOwner.getCivilizationDescription(0)))

            newPlayerIdx = -1
            # Overwrite a civ that never managed to found a city (probably a previous rebel)
            for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                if( (not gc.getPlayer(i).isAlive()) and (not gc.getPlayer(i).isFoundedFirstCity()) and (gc.getPlayer(i).getNumUnits() == 0) ) :
                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Dead civ %d has founded city? %d"%(i,gc.getPlayer(i).isFoundedFirstCity()))
                    newPlayerIdx = i
                    break

            if( newPlayerIdx < 0 ) :
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - No 'fake' civs, looking for empty slots")
                # Put in empty slot
                for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                    if( (not gc.getPlayer(i).isEverAlive()) and (gc.getPlayer(i).getNumUnits() == 0) ) :
                        newPlayerIdx = i
                        break

            if( newPlayerIdx < 0 ) :
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - No available slots (%d), spawning as Barbarians"%(newPlayerIdx))
                return gc.getPlayer(self.iBarbariansID)

            # Don't reincarnate as either of these
            iMinor = CvUtil.findInfoTypeNum(gc.getCivilizationInfo,gc.getNumCivilizationInfos(),'CIVILIZATION_MINOR')
            iBarbarian = CvUtil.findInfoTypeNum(gc.getCivilizationInfo,gc.getNumCivilizationInfos(),'CIVILIZATION_BARBARIAN')
            # Civs not currently in the game
            availableCivs = list()
            # Civs with similar style to cultOwner, if they exist
            similarStyleCivs = list()
            for civType in range(0,gc.getNumCivilizationInfos()) :
                if( not civType == iBarbarian ) :
                    if( not civType == iMinor ) :
                        taken = False
                        for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                            if( civType == gc.getPlayer(i).getCivilizationType() ) :
                                if( gc.getPlayer(i).isAlive() or gc.getPlayer(i).isFoundedFirstCity() ) :
                                    taken = True
                                    break
                        if( not taken ) :
                            availableCivs.append(civType)
                            if( not cultOwner == None ) :
                                if( gc.getCivilizationInfo( cultOwner.getCivilizationType() ).getArtStyleType() == gc.getCivilizationInfo(civType).getArtStyleType() ) :
                                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Potential similar style civ: %s"%(gc.getCivilizationInfo(civType).getShortDescription(0)))
                                    similarStyleCivs.append(civType)

            if( len(availableCivs) < 1 ) :
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - No available civs, spawning as Barbarians")
                return gc.getPlayer(self.iBarbariansID)

            if( len(similarStyleCivs) > 0 ) :
                # Create a civ using this style
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Creating a similar style civ to %s"%(cultOwner.getCivilizationDescription(0)))
                availableCivs = similarStyleCivs

            newCivIdx = availableCivs[game.getSorenRandNum(len(availableCivs),'Rebel: pick unused civ type')]

            leaderList = list()
            for leaderType in range(0,gc.getNumLeaderHeadInfos()) :
                if( gc.getCivilizationInfo(newCivIdx).isLeaders(leaderType) ) :
                    leaderList.append(leaderType)

            if( len(leaderList) < 1 ) :
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Unexpected lack of possible leaders, spawning as Barbarians")
                return gc.getPlayer(self.iBarbariansID)
            
            newLeaderIdx = leaderList[game.getSorenRandNum(len(leaderList),'Rebel: pick leader')]

            success = game.createNewPlayer( newPlayerIdx, newCivIdx, newLeaderIdx, -1, False, 1, 1 )

            if( not success ) :
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - New civ creation failed, spawning as Barbarians")
                return gc.getPlayer(self.iBarbariansID)

            pRebel = gc.getPlayer(newPlayerIdx)

            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Rebel civ %s spawned in slot %d"%(pRebel.getCivilizationDescription(0),newPlayerIdx))

            if( giveTechs ) :
                self.giveRebelTechs( pRebel, owner )

            return pRebel

        def giveRebelTechs( self, toPlayer, fromPlayer ) :

            # Give all techs known by from Player, except a few of the most expensive
            knownTechs = list()
            mostExpensive = list()
            minMostExpensive = 0
            numMostExpensive = 2 + self.getRand(3)

            fromPlayerTeam = gc.getTeam( fromPlayer.getID() )
            
            for techID in range(0,gc.getNumTechInfos()) :
                if( fromPlayerTeam.isHasTech( techID ) ) :
                    knownTechs.append( techID )
                    if( gc.getTechInfo( techID ).getResearchCost() > minMostExpensive ) :
                        if( len(mostExpensive) < numMostExpensive ) :
                            mostExpensive.append( techID )
                        else :
                            for j in range(0,len(mostExpensive)) :
                                if( gc.getTechInfo( mostExpensive[j] ).getResearchCost() == minMostExpensive ) :
                                    mostExpensive.remove( mostExpensive[j] )
                                    break
                            mostExpensive.append( techID )
                            minMostExpensive = gc.getTechInfo( mostExpensive[0] ).getResearchCost()
                            for j in range(0,len(mostExpensive)) :
                                if( gc.getTechInfo( mostExpensive[j] ).getResearchCost() < minMostExpensive ) :
                                    minMostExpensive = gc.getTechInfo( mostExpensive[j] ).getResearchCost()

            toPlayerTeam = gc.getTeam( toPlayer.getID() )
            for techID in knownTechs :
                if( not techID in mostExpensive ) :
                    toPlayerTeam.setHasTech(techID,True,toPlayer.getID(),False,False)
                    #if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Giving rebel %s"%(PyInfo.TechnologyInfo(techID).getDescription()))
                else :
                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - NOT giving rebel %s"%(PyInfo.TechnologyInfo(techID).getDescription()))


        def removeFloatingRebellions(self) :
            # Destroy all units for a rebel who is at peace, yet has no cities

            for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                player = gc.getPlayer(i)
                if( not player.isFoundedFirstCity() and player.getNumUnits() > 0 ) :
                    if( player.getNumUnits() == player.getNumMilitaryUnits() ) :
                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Player %d (%s) is a homeless rebel"%(player.getID(),player.getCivilizationDescription(0)))
                        if( gc.getTeam( player.getTeam() ).getAtWarCount( True ) == 0 ) :
                            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Player %d has lost thier cause, terminating rebel"%(player.getID()))
                            player.killUnits()
	
	def alterCitySpawnUnitCounters(self):
		
		# Loop through all players...
		for iPlayerLoop in range(gc.getMAX_CIV_PLAYERS()):
			
			pPlayer = gc.getPlayer(iPlayerLoop)
			
			# If this player is alive...
			if (pPlayer.isAlive()):
				
				apCityList = self.getCityList(pPlayer)
				
				# Loop through all this player's cities...
				for pCity in apCityList:
					
					iCityID = pCity.getID()
					# If this city's spawn counter is greater than 0...
					if (self.getCitySpawnUnitCounter(iPlayerLoop, pCity) > 0):
						# Reduce CitySpawnUnit counter by 1 in applicable cities
						self.changeCitySpawnUnitCounter(iPlayerLoop, iCityID, -1)

###########################################################################################
#################################### UTILITY FUNCTIONS ####################################
###########################################################################################
		
	def getRand(self, iNum):
		
		return CyGame().getSorenRandNum(iNum, "RebellionModScenario")
		
	def findUnitPlacementPlot(self, iPlayerID, iPlotX, iPlotY, bLand = True, bIncludePlot = True, bSameArea = True, iRange = 3, plotOwnerID = -1, bTryWider = False):
		
		aTHEPlot = []
		aiPossiblePlots = []
		iNumPossiblePlots = 0
		aiTempPlotHolder = []
		
		pThisPlayer = gc.getPlayer(iPlayerID)
		pThisTeam = gc.getTeam(pThisPlayer.getTeam())

		iBaseArea = CyMap().plot(iPlotX, iPlotY).area().getID()
		
		# Loop through plots in usable range
		for iXLoop in range(iPlotX - iRange, iPlotX + iRange):
			for iYLoop in range(iPlotY - iRange, iPlotY + iRange):
				
				# Map bounds
				if (iXLoop >= 0 and iXLoop < CyMap().getGridWidth() and iYLoop >= 0 and iYLoop < CyMap().getGridHeight()):
					
					# Check to see if center plot should be excluded
					if (bIncludePlot or (not bIncludePlot and (iXLoop != iPlotX) and (iYLoop != iPlotY))):
						
						pPlot = CyMap().plot(iXLoop, iYLoop)
						
						# Don't include cities
						if (pPlot.isCity() == false):
							
						    # Passable?
						    if (not pPlot.isImpassable()):

                                                        if( (plotOwnerID < 0 ) or (pPlot.getOwner() == plotOwnerID) ):
								
								# Check whether this plot type matches the type requested
								if ((bLand and not CyMap().plot(iXLoop, iYLoop).isWater()) or \
								    (not bLand and CyMap().plot(iXLoop, iYLoop).isWater())):

                                                                    if( (not bSameArea) or (CyMap().plot(iXLoop, iYLoop).area().getID() == iBaseArea) ) :
									aiTempPlotHolder = [iXLoop, iYLoop]
									aiPossiblePlots.append(aiTempPlotHolder)
									iNumPossiblePlots += 1
					
		if (iNumPossiblePlots == 0):
                        if( bTryWider and iRange < 10 ) :
                            return self.findUnitPlacementPlot(iPlayerID, iPlotX, iPlotY, bLand, bIncludePlot, bSameArea, iRange+1, plotOwnerID, bTryWider = False)
                        else :
                            return "Oh Snap"
		
		# Now that our list of possible plots is made, we pick one of them randomly and shove aside any enemy units already there
		iChosenPlot = self.getRand(iNumPossiblePlots - 1)
		aTHEPlot = aiPossiblePlots[iChosenPlot]
		
		# Check to see if there are any other units already on this plot
		pPlot = CyMap().plot(aTHEPlot[0], aTHEPlot[1])
		
		if (pPlot.getNumUnits() > 0):
			
			bEnemiesPresent = true
			
			while(pPlot.getNumUnits() > 0 and bEnemiesPresent):
				pLoopUnit = pPlot.getUnit(0)
				pUnitOwner = gc.getPlayer(pLoopUnit.getOwner())
				
				if (pUnitOwner):
					# Check to see if this unit belongs to a player that 'our' player is at war with
					if (pThisTeam.isAtWar(pUnitOwner.getTeam())):
						# Move this enemy unit
						self.moveUnitAside(pLoopUnit)
					else:
						# If it's a friendly, don't move it!
						bEnemiesPresent = false
		
		return aTHEPlot
		
	def moveUnitAside(self, pUnitToMove):
		
		iUnitX = pUnitToMove.getX()
		iUnitY = pUnitToMove.getY()
		
		# Don't look further than 10 plots away
		iMaxRange = 10
		iRing = 1
		
		aaiTestedList = [iUnitX, iUnitY]
		
		pPlayer = gc.getPlayer(pUnitToMove.getOwner())
		pTeam = gc.getTeam(pPlayer.getTeam())
		
		# Loop in rings around the unit's plot until we find another plot to move this unit to
		while (iRing < iMaxRange):
			
			aiTempList = []
			aaiToTestList = []
			iNumToTest = 0
			aaiValidPlotList = []
			iNumValid = 0
			
			# Find plots in this ring range
			for iX in range(-iRing, iRing + 1):
				for iY in range(-iRing, iRing + 1):
					aiTempList = [iX + iUnitX, iY + iUnitY]
					
					# Don't test this plot more than once
					if (aiTempList not in aaiTestedList):
						
						pPlot = CyMap().plot(aiTempList[0], aiTempList[1])
						bValidPlot = false
						
						if (pPlot.isWater() and pUnitToMove.getDomainType() == DomainTypes.DOMAIN_SEA):
							bValidPlot = true
						elif (pPlot.isWater() == false):
							if (pUnitToMove.getDomainType() == DomainTypes.DOMAIN_LAND or pUnitToMove.getDomainType() == DomainTypes.DOMAIN_AIR):
								bValidPlot = true
						
						# Make sure this unit can move to this plot type (no land units on water, etc.)
						if (bValidPlot):
							aaiToTestList.append(aiTempList)
							iNumToTest += 1
					
			# Now test these plots
			for iTestLoop in range(iNumToTest):
				
				pPlot = CyMap().plot(iUnitX, iUnitY)
				
				iX = aaiToTestList[iTestLoop][0]
				iY = aaiToTestList[iTestLoop][1]
				
				# If there are no enemy units here then add to the final valid plot list
				if (not self.isEnemyUnits(pTeam, iX, iY)):
					
					aaiValidPlotList.append(aaiToTestList[iTestLoop])
					iNumValid += 1
					
				# Add this plot to the tested list (to prevent it being looked at again) and look at the next in this ring
				aaiTestedList.append(aaiToTestList[iTestLoop])
				
			# This ring has been iterated through and placed in a final list - now pick a random plot from that list and exit the function
			
			if (iNumValid > 0):
				iChosenPlot = self.getRand(iNumValid - 1)
				
				iX = aaiValidPlotList[iChosenPlot][0]
				iY = aaiValidPlotList[iChosenPlot][1]
				
				pUnitToMove.setXY(iX, iY)
				
				return
				
			# None of the plots in this ring are valid, so time to try the next ring
			
			iRing += 1
		
	def isEnemyUnits(self, pThisTeam, iPlotX, iPlotY):
		
		# Check to see if there are any other units already on this plot
		pPlot = CyMap().plot(iPlotX, iPlotY)
		
		if (pPlot.getNumUnits() > 0):
			
			for iUnitLoop in range(pPlot.getNumUnits()):
				
				pLoopUnit = pPlot.getUnit(iUnitLoop)
				
				pUnitOwner = gc.getPlayer(pLoopUnit.getOwner())
				
				# Check to see if this unit belongs to a player that 'our' player is at war with
				if (pThisTeam.isAtWar(pUnitOwner.getTeam())):
					return true
		
		return false
	
	def getCityList(self, pPlayer):
		iPlayerID = pPlayer.getID()
		
		lCity = []
		(loopCity, iter) = pPlayer.firstCity(false)
		while(loopCity):
			cityOwner = loopCity.getOwner()
			if ( not loopCity.isNone() and loopCity.getOwner() == iPlayerID ): #only valid cities
				pCity = pPlayer.getCity(loopCity.getID())
#				city = PyCity( self.getID(), loopCity.getID() )
				lCity.append(pCity)
			(loopCity, iter) = pPlayer.nextCity(iter, false)
		return lCity
		
	def addPopup(self, szText, szTitle):
		
#		szTitle = self.szGameDate = CyGameTextMgr().getTimeStr(CyGame().getGameTurn(), false)
		
		popup = PyPopup.PyPopup(-1)
		popup.setHeaderString(szTitle)
		popup.setBodyString(szText)
		popup.launch(true, PopupStates.POPUPSTATE_QUEUED)



###########################################################################################
####################################### SCRIPT DATA #######################################
###########################################################################################
	
	def initCity(self, pCity):
		
		# Parameter 0 is City Spawn Unit Counter Max
		# 1 is rebel player ID
		szScriptData = [self.iCitySpawnUnitCounterMax, -1]
		
		# Save Script Data - SpawnUnitCounter
		pCity.setScriptData(pickle.dumps(szScriptData))
	
	def getCitySpawnUnitCounter(self, iPlayer, pCity):
		
		pPlayer = gc.getPlayer(iPlayer)
#		pCity = pPlayer.getCity(iCity)  
		
		# Load Script Data - SpawnUnitCounter
		try :
                    szScriptData = pickle.loads(pCity.getScriptData())
                    iSpawnUnitCounter = szScriptData[self.iCitySpawnScriptID]
                except (EOFError, IndexError):
                    self.initCity(pCity)
                    szScriptData = pickle.loads(pCity.getScriptData())
                    iSpawnUnitCounter = szScriptData[self.iCitySpawnScriptID]

		#print("Rebel - Loading spawn unit counter, val: %d"%(iSpawnUnitCounter))
		
		return iSpawnUnitCounter
		
	def setCitySpawnUnitCounter(self, iPlayer, pCity, iValue):
		
		pPlayer = gc.getPlayer(iPlayer)
#		pCity = pPlayer.getCity(iCity)
		
		# Load Script Data - SpawnUnitCounter
		try :
                    szScriptData = pickle.loads(pCity.getScriptData())
                    szScriptData[self.iCitySpawnScriptID] = iValue
                except (EOFError, IndexError):
                    self.initCity(pCity)
                    szScriptData = pickle.loads(pCity.getScriptData())
                    szScriptData[self.iCitySpawnScriptID] = iValue
		
		# Save Script Data - SpawnUnitCounter
		pCity.setScriptData(pickle.dumps(szScriptData))
		
	def changeCitySpawnUnitCounter(self, iPlayer, iCity, iChange):
		
		pPlayer = gc.getPlayer(iPlayer)
		pCity = pPlayer.getCity(iCity)
		
		# Load Script Data - SpawnUnitCounter
		try :
                    szScriptData = pickle.loads(pCity.getScriptData())
                    iSpawnUnitCounter = szScriptData[self.iCitySpawnScriptID]
                except (EOFError, IndexError):
                    self.initCity(pCity)
                    szScriptData = pickle.loads(pCity.getScriptData())
                    iSpawnUnitCounter = szScriptData[self.iCitySpawnScriptID]
		
		iSpawnUnitCounter += iChange
		
		# Save Script Data - SpawnUnitCounter
		szScriptData[self.iCitySpawnScriptID] = iSpawnUnitCounter
		pCity.setScriptData(pickle.dumps(szScriptData))

	def getCityRebelPlayer( self, pCity ) :
                # Load Script Data - RebelPlayer
		try :
                    szScriptData = pickle.loads(pCity.getScriptData())
                    #print("Rebel - get rebel player from city load returns list of length %d"%(len(szScriptData)))
                    #print szScriptData                          
                    iRebelPlayer = szScriptData[self.iCityRebelPlayerID]
                except (EOFError, IndexError):
                    print("Rebel - get rebel player from city failed, doing reinit")
                    self.initCity(pCity)
                    szScriptData = pickle.loads(pCity.getScriptData())
                    iRebelPlayer = szScriptData[self.iCityRebelPlayerID]

                print("Rebel - Rebel player for %s is %d"%(pCity.getName(),iRebelPlayer))
                return iRebelPlayer

	def setCityRebelPlayer( self, iPlayer, pCity ) :
                # Load Script Data - RebelPlayer
		try :
                    szScriptData = pickle.loads(pCity.getScriptData())
                    szScriptData[self.iCityRebelPlayerID] = iPlayer
                except (EOFError, IndexError):
                    self.initCity(pCity)
                    szScriptData = pickle.loads(pCity.getScriptData())
                    szScriptData[self.iCityRebelPlayerID] = iPlayer
                
                pCity.setScriptData(pickle.dumps(szScriptData))

                print("Rebel - Rebel player for %s set to %d"%(pCity.getName(),iPlayer))
                #print("Rebel - Attempted readback from %s yields %d"%(pCity.getName(),self.getCityRebelPlayer(pCity)))
